﻿using KPGBase.Excercises.KPGGameOfLife;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace KPGBase.Excercises
{
    internal class Excercise03 : BasicExcercise
    {
        GameOfLife gol;
        int cellsize;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="canvas"> Canvas - bitmap image </param>
        public Excercise03(Bitmap canvas) : base(canvas)
        {
            gol = new GameOfLife();
        }

        /// <summary>
        /// Generate next step of the game of life and draw it
        /// </summary>
        /// <param name="m"> Number of rows </param>
        /// <param name="n"> Number of columns </param>
        /// <param name="cellsize"> Size of one cell </param>
        public void NextLifeConfig()
        {
            // next step of life
            gol.NextStep();
            // draw all cells
            DrawLife(gol, cellsize);
        }

        /// <summary>
        /// Generate some life configuration and draw it
        /// </summary>
        /// <param name="m"> Number of rows </param>
        /// <param name="n"> Number of columns </param>
        /// <param name="cellsize"> Size of one cell </param>
        public void GenerateLife(int m, int n, int cellsize)
        {
            this.cellsize = cellsize;
            // create a new configuration
            gol.GenerateRandLife(m, n);
            // draw all cells
            DrawLife(gol, cellsize);
        }

        /// <summary>
        /// Draw the current configuration
        /// </summary>
        /// <param name="gol"> Game of life to draw </param>
        /// <param name="cellsize"> Size of one cell </param>
        private void DrawLife(GameOfLife gol, int cellsize)
        {
            ClearImage();
            Pen p = new Pen(Color.Gray, 2);
            Brush b = new SolidBrush(Color.Green);

            // go through all cells
            for (int i = 0; i < gol.cells.GetLength(0); i++)
            {
                for (int j = 0; j < gol.cells.GetLength(1); j++)
                {

                    Point pNW = new Point(cellsize + j * cellsize, cellsize + i * cellsize);
                    Rectangle cellRec = new Rectangle(pNW, new Size(cellsize, cellsize));
                    if (gol.cells[i, j].alive)
                        g.FillRectangle(b, cellRec);

                    g.DrawRectangle(p, cellRec);
                }
            }
        }

    }
}
