﻿using System.Drawing.Drawing2D;

namespace KPGBase.Excercises
{
    public class Point3D
    {
        public float x, y, z;

        public Point3D() : this(0, 0, 0)
        {
        }

        public Point3D(float x, float y, float z)
        {
            this.x = x;
            this.y = y;
            this.z = z;
        }
    }

    internal class ProjectionMatrix
    {
        public float[,] projectionMatrix = new float[4, 4]; // Matice projekce

        /// <summary>
        /// Orthographic projection
        /// </summary>
        public ProjectionMatrix()
        {
            /* Set projection matrix to
             * 
             *      1  0  0  0
             *      0  1  0  0
             *      0  0  0  0
             *      0  0  0  1
             */

            throw new NotImplementedException();
        }

        /// <summary>
        /// Pseudo-projection matrix (Z-divide)
        ///     - camera in (0,0,0), screen in z=d
        /// </summary>
        /// <param name="distance"> Distance </param>
        public ProjectionMatrix(float distance)
        {
            /* Set projection matrix to
             * 
             *      1  0   0  0
             *      0  1   0  0
             *      0  0   1  0
             *      0  0  1/d 0
             *      
             *  d = distance
             */

            throw new NotImplementedException();
        }

        // Perspektivni projekce
        public ProjectionMatrix(float zNear, float zFar, float viewportWidth, float viewportHeight)
        {
            /* Nastavit matici ortograficke projekce na
             * 
             *      w  0  0     0
             *      0  h  0     0
             *      0  0  -Q1  Q2
             *      0  0  -1     0
             *      
             *      w = (2 * zNear) / width
             *      h = (2 * zNear) / height
             *      Q1 = - (zFar + zNear) / (zFar - zNear)
             *      Q2 = (2 * zNear * zFar) / (zNear - zFar)
             *      
             */

            throw new NotImplementedException();
        }

        /// <summary>
        /// Set identity matrix
        /// </summary>
        public void Identity()
        {
            for (int i = 0; i < projectionMatrix.GetLength(0); i++)
            {
                for (int j = 0; j < projectionMatrix.GetLength(1); j++)
                {
                    if (i == j) projectionMatrix[i, j] = 1;
                    else projectionMatrix[i, j] = 0;
                }
            }
        }

        /// <summary>
        /// Transform point p by projection matrix
        /// </summary>
        /// <param name="p"> 3D point </param>
        /// <returns> Projected point </returns>
        public Point3D Multiply(Point3D p)
        {
            Point3D result = new Point3D(0, 0, 0);
            // Homogenous coordinate
            float w = projectionMatrix[3, 0] * p.x + projectionMatrix[3, 1] * p.y + projectionMatrix[3, 2] * p.z + projectionMatrix[3, 3] * 1;

            // Transformed point
            result.x = (projectionMatrix[0, 0] * p.x + projectionMatrix[0, 1] * p.y + projectionMatrix[0, 2] * p.z + projectionMatrix[0, 3] * 1) / w;
            result.y = (projectionMatrix[1, 0] * p.x + projectionMatrix[1, 1] * p.y + projectionMatrix[1, 2] * p.z + projectionMatrix[1, 3] * 1) / w;
            result.z = (projectionMatrix[2, 0] * p.x + projectionMatrix[2, 1] * p.y + projectionMatrix[2, 2] * p.z + projectionMatrix[2, 3] * 1) / w;

            return result;
        }
    }

    internal class Excercise11 : BasicExcercise
    {
        // Type of projection matrix, 1 - ortho, 2 - z-divide, 3 - perspective
        private int projectionType = 3;
        
        // Viewport
        private float viewportWidth = 0.25f;
        private float viewportHeight = 0.25f;
        private float zNear = 1;  
        private float zFar = 100;

        // Scale
        private float scale = 25;

        // All points to be drawn
        List<Point3D> vertexBuffer = new List<Point3D>();
        // List of triangles
        List<int> indexBuffer = new List<int>();
        // Rotation axis
        Point3D rotationAxis = new Point3D(0, 0, 0);
        // Projection matrix
        ProjectionMatrix projection = null;

        /// <summary>
        /// Constructor
        /// </summary>
        public Excercise11(Bitmap canvas) : base(canvas)
        {
            // Origin of drawn section is in the middle of the canvas
            Matrix origin = new Matrix();
            origin.Translate(canvas.Width / 2, canvas.Height / 2);
            g.Transform = origin;
        }

        /// <summary>
        /// Prepare model before launching animation
        /// </summary>
        public void PrepareModel()
        {
            if (vertexBuffer == null || vertexBuffer.Count == 0)
                CreateBox();
        }

        /// <summary>
        /// Draw next step of animation
        /// </summary>
        public void AnimationStep()
        {
            // Create projection matrix
            if (projectionType == 1) projection = new ProjectionMatrix(); // ortho
            else if (projectionType == 2) projection = new ProjectionMatrix(9); // z-divide
            else projection = new ProjectionMatrix(zNear, zFar, viewportWidth, viewportHeight); // perspective

            // Rotate each vertex around rotation axis by angle increment
            double angle = 0.001;
            for (int i = 0; i < vertexBuffer.Count; i++)
                vertexBuffer[i] = RotateVector(vertexBuffer[i], rotationAxis, angle);

            // Draw object
            DrawObject(vertexBuffer, indexBuffer);
        }

        /// <summary>
        /// Rotate point p around y-axis located at point rotationAxis by angle
        /// </summary>
        /// <param name="p"> Point to rotate </param>
        /// <param name="rotationAxis"> Location of axis </param>
        /// <param name="angle"> Angle </param>
        /// <returns> Transformed point </returns>
        private Point3D RotateVector(Point3D p, Point3D rotationAxis, double angle)
        {
            // Rotate point p around y-axis located at point rotationAxis by angle
            // - move point to the rotationAxis location -> so that rotation axis passes through the point
            // - rotate around y (rotation matrix in slides)
            // - move point back

            throw new NotImplementedException();
        }

        private void DrawObject(List<Point3D> vert, List<int> ind)
        {
            ClearImage();

            // Draw object with vertices stored in list vert and triangles stored in ind
            // For each triangle
            // - vertex indices stored in list ind (each triangle has 3 vertex indices)
            // - compute projected location of all vertices
            //      - newPos = projection.multiply(oldPos)
            // - create 2D point
            //      - get x and y from newPos
            //      - multiply each coordinate by scale
            // - draw the edges of the triangle
            //
            // !! ATTENTION !! check if the point's z-coordinate is closer than zFar
            // To check if zFar cut is working try drawing ellipses for all points, red for points that are too far and white for the ones in frustrum

            throw new NotImplementedException();
        }

        /// <summary>
        /// Create simple box
        /// </summary>
        private void CreateBox()
        {
            // Box vertices
            Point3D a = new Point3D(3, 3, 7);
            Point3D b = new Point3D(3, 3, 13);
            Point3D c = new Point3D(3, -3, 13);
            Point3D d = new Point3D(3, -3, 7);

            Point3D e = new Point3D(-3, 3, 7);
            Point3D f = new Point3D(-3, 3, 13);
            Point3D g = new Point3D(-3, -3, 13);
            Point3D h = new Point3D(-3, -3, 7);

            // Add to vertex buffer
            vertexBuffer.Add(a);
            vertexBuffer.Add(b);
            vertexBuffer.Add(c);
            vertexBuffer.Add(d);

            vertexBuffer.Add(e);
            vertexBuffer.Add(f);
            vertexBuffer.Add(g);
            vertexBuffer.Add(h);

            // Create triangles
            // right side
            indexBuffer.Add(0);
            indexBuffer.Add(1);
            indexBuffer.Add(2);
            indexBuffer.Add(0);
            indexBuffer.Add(2);
            indexBuffer.Add(3);

            // front side
            indexBuffer.Add(4);
            indexBuffer.Add(0);
            indexBuffer.Add(3);
            indexBuffer.Add(4);
            indexBuffer.Add(3);
            indexBuffer.Add(7);

            // left side
            indexBuffer.Add(5);
            indexBuffer.Add(7);
            indexBuffer.Add(4);
            indexBuffer.Add(5);
            indexBuffer.Add(7);
            indexBuffer.Add(6);

            // back side
            indexBuffer.Add(5);
            indexBuffer.Add(1);
            indexBuffer.Add(6);
            indexBuffer.Add(1);
            indexBuffer.Add(6);
            indexBuffer.Add(2);

            // bottom side
            indexBuffer.Add(2);
            indexBuffer.Add(3);
            indexBuffer.Add(7);
            indexBuffer.Add(2);
            indexBuffer.Add(7);
            indexBuffer.Add(6);

            // top side
            indexBuffer.Add(0);
            indexBuffer.Add(1);
            indexBuffer.Add(4);
            indexBuffer.Add(1);
            indexBuffer.Add(4);
            indexBuffer.Add(5);

            // Rotation axis location
            rotationAxis = new Point3D(0, 0, 10);
        }

        public float Scale
        {
            get { return this.scale; }
            set { this.scale = value; }
        }

        public int ProjectionType
        {
            get { return this.projectionType; }
            set { this.projectionType = value; }
        }

        public float ViewportWidth
        {
            get { return this.viewportWidth; }
            set { this.viewportWidth = value; }
        }

        public float ViewportHeight
        {
            get { return this.viewportHeight; }
            set { this.viewportHeight = value; }
        }

        public float ZNear
        {
            get { return this.zNear; }
            set { this.zNear = value; }
        }

        public float ZFar
        {
            get { return this.zFar; }
            set { this.zFar = value; }
        }
    }
    
}
