﻿namespace KPGBase.Excercises.KPGGameOfLife
{
    internal class GameOfLife
    {
        /// <summary> Grid with cells that will be used for the simulation </summary>
        public LifeCell[,] cells;
        /// <summary> Number of rows </summary>
        int m;
        /// <summary> Number of columns </summary>
        int n;

        /// <summary>
        /// Generate the starting configuration
        /// </summary>
        /// <param name="m"> Number of rows </param>
        /// <param name="n"> Number of columns </param>
        internal void GenerateRandLife(int m, int n)
        {
            this.m = m; this.n = n;

            cells = new LifeCell[m, n];
            for (int i = 0; i < cells.GetLength(0); i++)
            {
                for (int j = 0; j < cells.GetLength(1); j++)
                {
                    cells[i, j] = new LifeCell(i, j);

                }
            }

            // TODO implement random initialization of all cells

            // this is a simple glider
            cells[0, 1].alive = true;
            cells[1, 2].alive = true;
            cells[2, 0].alive = true;
            cells[2, 1].alive = true;
            cells[2, 2].alive = true;
        }

        /// <summary>
        /// Method for the simulation of Conway's game of life 
        /// - apply the rules on all of the cells of the current configuration
        /// - create a new configuration that can be visualized
        /// </summary>
        internal void NextStep()
        {
            LifeCell[,] newCells = new LifeCell[m, n];

            for (int i = 0; i < cells.GetLength(0); i++)
            {
                for (int j = 0; j < cells.GetLength(1); j++)
                {
                    LifeCell curr = cells[i, j];
                    newCells[i, j] = new LifeCell(i, j);

                    int nns = GetNumAliveNns(cells[i, j]);
                    if (curr.alive && (nns == 2 || nns == 3))
                        newCells[i, j].alive = true;
                    if (!curr.alive && nns == 3)
                        newCells[i, j].alive = true;
                }
            }

            cells = newCells;
        }

        /// <summary>
        /// Get number of living neighbours of the cell c
        /// </summary>
        /// <param name="c"> Cell </param>
        /// <returns> Number of living neighbors </returns>
        private int GetNumAliveNns(LifeCell c)
        {
            int m = cells.GetLength(0);
            int n = cells.GetLength(1);

            int xn = c.row - 1; // north neighbor row index
            int xs = (c.row + 1) % m; // south neighbor row index
            int ye = (c.col + 1) % n; // east neighbor column index
            int yw = c.col - 1; // west neighbor column index

            if (xn < 0)
                xn = m - 1;
            if (yw < 0)
                yw = n - 1;

            // For each index test if it is a valid index for the grid, and test if the neighbor is alive
            int alive = 0;

            // north neighbors
            if (xn >= 0)
            {
                if (cells[xn, c.col].alive)
                    alive++;

                if (ye < cells.GetLength(1) && cells[xn, ye].alive)
                    alive++;
                if (yw >= 0 && cells[xn, yw].alive)
                    alive++;
            }

            // south neighbors
            if (xs < cells.GetLength(0))
            {
                if (cells[xs, c.col].alive)
                    alive++;

                if (ye < cells.GetLength(1) && cells[xs, ye].alive)
                    alive++;
                if (yw >= 0 && cells[xs, yw].alive)
                    alive++;
            }

            if (ye < cells.GetLength(1) && cells[c.row, ye].alive)
                alive++;
            if (yw >= 0 && cells[c.row, yw].alive)
                alive++;

            return alive;
        }

    }
}